


/**************

 Green LED:
   Blinks when an interest, reinforcement for myself is recvd

 Red LED:
   Blinks when an interest, High data rate packet is recvd for myself

 Yellow LED:
   Blinks when an interest, Low Data Rate packet is recvd   

**************/


/* Todo

1. Create a timeout to flush cache completely, so as to avoid stale data.
2. Create a timeout to flush gradients which are in the INTEREST_STATE;
3. Create different structures for data and reinf messages.

a) Connect one or few nodes to the UART so that query can be specified 
   over the UART. Make this a separate access point into the diffusion
   network.
b) Make Diffusion multi-sink multi-source.
c) Do a Agg compute and route towards sink.

QUERY: Give me temperature averages from region X and region Y.

*/
   


#include "tos.h"
#include "DIFFUSION.h"

#define REMOTE_DIFFUSION_PKT 1
#define LOCAL_AGENT_PKT 2

#define NULL_STATE 0
#define PENDING_STATE 1

#define INTEREST_TYPE 102
#define REINFORCEMENT_TYPE 103
#define LOW_DATA_RATE_TYPE 104
#define HIGH_DATA_RATE_TYPE 105

#define INTEREST_STATE 2
#define REINFORCED_STATE 3
#define LOW_DATA_RATE_STATE 4
#define HIGH_DATA_RATE_STATE 5

#define INTEREST_TIMEOUT 40 //10 secs
#define REINF_TIMEOUT 16 // 4 secs

#define MAX_CACHE_SIZE 10
#define MAX_GRADIENTS 8

#define MAX_SINKS_PER_DATA_TYPE 1
#define MAX_SRCS_PER_DATA_TYPE 5

#define MAX_NBR_LIST 15

#define SENTINEL 0xff

#define MAX_HDR_PACKETS 200
#define MAX_LDR_PACKETS 50

#define HDR 2
#define LDR 8

#define RPS(i,j) VAR(gradient_list[i]).rps[j]
#define ARRIVAL_ORDER(i,j) VAR(gradient_list[i]).arrival_order[j/2]

#define fflush(stdout);

extern const char TOS_LOCAL_ADDRESS;

typedef struct{
  unsigned char data_type;
  unsigned char pkt_seq;
  unsigned char pkt_src;
  unsigned char data_src;
  unsigned char data_sink;
  unsigned char prev_hop;
  unsigned char next_hop;
  unsigned char reinf_pkt_src;
  unsigned char reinf_pkt_seq;
  unsigned char result;
} query_msg;

typedef struct {
  unsigned char data_src;
  unsigned char reinf_nbr;
} reinforcement_per_src;

typedef struct{
  // data_type & data_sink uniquely identify query
  unsigned char bit_fields;
  unsigned char data_sink;
  unsigned char interest_seq;
  // rate_counter is timeout for interest???
  unsigned char rate;                // Rate at which data is requested
  unsigned char rate_counter;        // Local counter to timeout
  reinforcement_per_src rps[MAX_SRCS_PER_DATA_TYPE];
  unsigned char arrival_order[2]; // The arrival order points to an index in nbr_list
} gradient_entry;


typedef struct{
  unsigned char pkt_src;
  unsigned char pkt_seq;
  unsigned char prev_hop;
} cache_entry;

typedef struct{
  unsigned char port;
} register_entry;

#define TOS_FRAME_TYPE DIFFUSION_OBJ
TOS_FRAME_BEGIN(DIFFUSION_OBJ) {
  TOS_Msg buf;                  // buffer sent for outgoing packets
  TOS_MsgPtr sentMsg;           // pointer to outgoing packet saved (?)
  TOS_MsgPtr sentThroughBuffer; // pointer to buffer sent from connected component
  TOS_MsgPtr msg;
  char pending;                 // if there is a transmission pending
  char msg_buffer_pending;
  unsigned char cache_index;    // CacheIndex
  cache_entry cache[10];        // cache
  gradient_entry gradient_list[MAX_GRADIENTS]; // gradient list
  unsigned char my_location;    // my location (currently hardcoded)
  unsigned char max_seq;         // my seq no: updated for packets generated by me.
  char state;                   // state?
  char photo_data;              // photo data
  char photo_pending;           // state of photo sensor
  unsigned char nbrlist_index; 
  unsigned char nbrlist[MAX_NBR_LIST];   // List of neighbors currently seen.
  unsigned int shift_reg;
  unsigned char packets_sent;
  unsigned char port_reg[8];
  
  unsigned char led_state;
  
}
TOS_FRAME_END(DIFFUSION_OBJ);

static inline unsigned char GetState(unsigned char i) 
{return (VAR(gradient_list[i].bit_fields) & 0x07);}

static inline void SetState(unsigned char i, unsigned char x) 
{VAR(gradient_list[i]).bit_fields |= x;}

static inline unsigned char GetDataType(unsigned char i) 
{return ((VAR(gradient_list[i].bit_fields) & 0x38) >> 3);}

static inline void SetDataType(unsigned char i, unsigned char x)
{VAR(gradient_list[i]).bit_fields |= (x << 3);}

static inline unsigned char GetArrivalCount(unsigned char i) 
{return((VAR(gradient_list[i].bit_fields) & 0xc0) >> 6);}

static inline void SetArrivalCount(unsigned char i, unsigned char x) 
{VAR(gradient_list[i]).bit_fields |= (x << 6);}

static inline unsigned char IsSource(unsigned char data_type) {return 0;}

static inline TOS_MsgPtr TOS_EVENT(DIFFUSION_NULL_FUNC)(TOS_MsgPtr msg, char i){return NULL;};

/* Init Routine
   Initialize gradient_list
                 state = NULL
   Initialize cache
   Initialize my_location
   Initialize max_seq = 0
   Initialize clock, rfm...
*/
char TOS_COMMAND(DIFFUSION_INIT)(){
  unsigned char i;

  VAR(led_state)=0;

  VAR(state) = 0;
  VAR(pending) = 0;
  VAR(msg_buffer_pending) = 0;
  VAR(cache_index) = 0;

  for (i=0;i<10;i++) {
    VAR(cache[i]).pkt_src = SENTINEL;
    VAR(cache[i]).pkt_seq = SENTINEL;
  }

  for (i=0;i<8;i++)
    VAR(port_reg[i]) = SENTINEL;
    //    VAR(port_reg[i]).port = SENTINEL;

  for (i=0;i<MAX_GRADIENTS;i++)
    ClearGradientState(i,'I');

  VAR(msg) = &VAR(buf);

  VAR(max_seq) = 0;

  VAR(my_location) = TOS_LOCAL_ADDRESS;
  VAR(shift_reg) = 119*VAR(my_location);
  VAR(packets_sent) = 0;
  TOS_CALL_COMMAND(DIFFUSION_SUB_INIT)();
  return 1;
}

char TOS_COMMAND(DIFFUSION_START)(){
  return 1;
}

char TOS_COMMAND(DIFFUSION_REGISTER)(unsigned char port, unsigned char data_type) {
  //  VAR(port_reg[data_type]).port = port;
  VAR(port_reg[data_type]) = port;
  return port;
}


TOS_MsgPtr TOS_COMMAND(DIFFUSION_GET_MSG_BUFFER)() {
  printf("In DIFFUSION_GET_MSG_BUFFER\n");fflush(stdout);
  if (!VAR(msg_buffer_pending)) {
    VAR(msg_buffer_pending) = 1;
    return VAR(msg);
  }
  return NULL;
}
 
char TOS_COMMAND(DIFFUSION_RELEASE_MSG_BUFFER)(TOS_MsgPtr msg) {
  if (VAR(msg_buffer_pending)) {
    VAR(msg_buffer_pending) = 0;
    VAR(msg) = msg;
    return 1;
  }
  return 0;
}


/* Send completion event
   Determine if this event was through Diffusion.
   If so, send DONE to all agents so that they can do needful.
   
   sentMsg matches => packet has been sent by Diffusion itself or an Agent over Diffusion.
     If Diffusion itself, then pending is reset => VAR(msg) becomes free.
     If Agent over Diffusion has requested buffer, its responsibility
     to release buffer.
   Reset pending so that VAR(msg) can be used.

   If there exist pending messages to be sent, then send them

*/
char TOS_EVENT(DIFFUSION_SUB_MSG_SEND_DONE)(TOS_MsgPtr sentBuffer){
  char sig=0;

  // If the event is ours, signal event
  
  printf("DIFFUSION_SUB_MSG_SEND_DONE\n");fflush(stdout);
  
  if (VAR(pending)) {
    printf("VAR(pending)=1\n");
    if (sentBuffer == VAR(sentMsg)) {
      printf("sentBuffer == VAR(sentMsg)\n");
      VAR(pending) = 0;
      VAR(msg_buffer_pending) = 0;
      sig = 1;
    } else {
      if (sentBuffer == VAR(sentThroughBuffer)) {
	printf("sentBuffer == VAR(sentThroughBuffer)\n");
	VAR(pending) = 0;
	VAR(sentThroughBuffer) = NULL;
	TOS_SIGNAL_EVENT(DIFFUSION_MSG_SEND_DONE)(sentBuffer);
	sig = 1;
      } else sig=0;
    }
  }
  return 1;
}


/***********************

Choose a random delay between 0 and limit clock ticks.

**********************/

int RandomDelay(unsigned char limit) {
  char bit;
  bit = (VAR(shift_reg) & 0x2) >> 1;
  bit ^= ((VAR(shift_reg) & 0x4000) >> 14);
  bit ^= ((VAR(shift_reg) & 0x8000) >> 15);
  VAR(shift_reg) >>=1;
  if (bit & 0x1)
    VAR(shift_reg) |= 0x8000;
  return ((VAR(shift_reg) & limit));
}


/* Cache check:
     Check for pkt_seq, pkt_src
     The assumption is that the pkt_src does not recycle sequence numbers.
*/
char InCache(unsigned char pkt_seq, unsigned char pkt_src, unsigned char prev_hop) {
  unsigned char i;

  for (i=0;i<10;i++) if ((VAR(cache[i]).pkt_seq==pkt_seq) && (VAR(cache[i]).pkt_src==pkt_src))
    return 1;

  VAR(cache[VAR(cache_index)]).pkt_seq = pkt_seq;
  VAR(cache[VAR(cache_index)]).pkt_src = pkt_src;
  VAR(cache[VAR(cache_index)]).prev_hop = prev_hop;

  if (++VAR(cache_index) >9) VAR(cache_index)=0;
  return 0;
}

unsigned char GetCacheEntry(unsigned char pkt_seq, unsigned char pkt_src) {
  unsigned char i=0;
  
  for (i=0;i<10;i++) if ((VAR(cache[i]).pkt_seq==pkt_seq) && (VAR(cache[i]).pkt_src==pkt_src))
    return (VAR(cache[i]).prev_hop);

  return SENTINEL;
}

/* 
   Searches gradient_list for entry corresponding to data_type/data_sink
   If data_type/data_sink does not exist, then create a corresponding entry.
*/
unsigned char GetGradientEntry(unsigned char data_type, unsigned char data_sink) {
  unsigned char i=0,empty=SENTINEL;

  for (i=0;i<5;i++) {
    printf("GradientList: %d, %d\n",GetDataType(i),VAR(gradient_list[i]).data_sink);fflush(stdout);
    if (VAR(gradient_list[i]).data_sink==data_sink && GetDataType(i)==data_type) return i;
    if (VAR(gradient_list[i]).data_sink==SENTINEL) empty = i;
  }

  printf("GetGradientEntry(%d, %d): Returned %d\n",data_type, data_sink, empty);fflush(stdout);

  // If entry does not exist create entry if slot available
  if (empty != SENTINEL) {
    SetDataType(empty, data_type);
    VAR(gradient_list[empty]).data_sink == data_sink;
    return empty;
  }
  return SENTINEL;
}


 /* Called
    When interest is not refreshed
      everything corresponding to the interest is purged

    When a new interest_seq is recvd
      everything corresponding to the prev interest is purged
      
 */

void ClearGradientState(unsigned char i, char type) {

  int k=0;

  if (type == 'I') {
    
    VAR(gradient_list[i]).data_sink = SENTINEL;
    VAR(gradient_list[i]).interest_seq = SENTINEL;

    for (k=0;k<MAX_SRCS_PER_DATA_TYPE;k++) {
      VAR(gradient_list[i]).rps[k].data_src = SENTINEL;
      VAR(gradient_list[i]).rps[k].reinf_nbr = SENTINEL;
    }
    VAR(gradient_list[i]).bit_fields = 0x00; // Set all bit fields to 0
    VAR(gradient_list[i]).rate = 0x00;
    VAR(gradient_list[i]).rate_counter = 0x00;
    VAR(gradient_list[i]).arrival_order[0] = VAR(gradient_list[i]).arrival_order[1] = 0x00;
    return;
  }

  // If new reinforcement is recvd from data_sink,
  // Or if reinforcement timeout occurs,
  // Reinforcement list is emptied of data_sink entry.
  if (type == 'R') {
  }

}

/* Add to gradient list when INTEREST packet arrives
   Match gradient with data_type/data_sink pair.
     => only one query can be associated with a data_type/data_sink pair
   Gradient list has the order of packet arrival for interest packets

   Note: CANNOT handle multiple data_sinks
   
*/
unsigned char AddGradient(unsigned char data_type, unsigned char interest_seq, unsigned char data_sink, unsigned char prev_hop) {
  unsigned char i=0;
  
  i = GetGradientEntry(data_type, data_sink);

  // If gradientlist is full and the data_type/data_sink pair is new, return
  if (i==SENTINEL) return SENTINEL;

  printf("AddGradient : %d\n",i);
  
  // If a new seq number is seen for interest
  // flush gradient[] list
  // This takes care of both new entries and new interests
  if (interest_seq != VAR(gradient_list[i]).interest_seq) {
    ClearGradientState(i,'I');
    VAR(gradient_list[i]).data_sink = data_sink;
    SetDataType(i, data_type);
    VAR(gradient_list[i]).interest_seq = interest_seq;
    AddArrival(i,prev_hop);
  } else   
    AddArrival(i,prev_hop);

  return i;

}


/*
  Searches Reinforced Nbr entries for corresponding data_src
*/

unsigned char GetReinfNbr(unsigned char i, unsigned char data_src)
{
  unsigned char j=0;

  for (j=0; j<MAX_SRCS_PER_DATA_TYPE; j++) 
    if (RPS(i,j).data_src == data_src) return RPS(i,j).reinf_nbr;

  // Src entry not found.
  return SENTINEL;
}

unsigned char AddReinfNbr(unsigned char i, unsigned char data_src, unsigned char reinf_nbr) {
  unsigned char j=0, empty=SENTINEL;

  for (j=0; j<MAX_SRCS_PER_DATA_TYPE; j++) {
    printf("RPS(%d,%d).data_src = %d\n",i,j,RPS(i,j).data_src);
    if (RPS(i,j).data_src == data_src) {RPS(i,j).reinf_nbr = reinf_nbr; return 1;}
    if (RPS(i,j).data_src == SENTINEL) empty = j;
  }

  if (empty != SENTINEL) {RPS(i,empty).data_src = data_src; RPS(i,empty).reinf_nbr = reinf_nbr;}
  return empty;
}

unsigned char AddArrival(unsigned char i, unsigned char prev_hop) {
  unsigned char j=0,k=0;

  if ((k = GetNbrListIndex(prev_hop)) == SENTINEL) return SENTINEL;

  if ((j=GetArrivalCount(i)) < 3)
      if (j%2)
	ARRIVAL_ORDER(i,j) |= (0x0f & k);
      else
	ARRIVAL_ORDER(i,j) |= (0xf0 & (k<<4));
  else return SENTINEL;
  return 1;
}

unsigned char AddToNbrList(unsigned char nbr) {

  if (VAR(nbrlist_index) < MAX_NBR_LIST)
    VAR(nbrlist[VAR(nbrlist_index)]) = nbr;
  else return SENTINEL;
  VAR(nbrlist_index)++;
  return 1;
}

unsigned char GetNbrListIndex(unsigned char nbr) {
  unsigned char i=0; 
     
  for (i=0; i<VAR(nbrlist_index); i++)
    if (VAR(nbrlist[i]) == nbr) return i;

  return SENTINEL;
}

/* 
   Agent wishes to send packet, say HDR, and uses DIFFUSION gradient tables.
*/
void TOS_COMMAND(DIFFUSION_PROCESS_AND_SEND_MSG)(TOS_MsgPtr msg, char pkt_type) {
  ProcessPacket(msg, pkt_type, LOCAL_AGENT_PKT);
}

/* 
   Agent wishes to send packet, say HDR, and does not wish to use DIFFUSION 
   gradient tables.
*/
char TOS_COMMAND(DIFFUSION_SEND_MSG)(TOS_MsgPtr msg, char pkt_type) {
  query_msg* q=(query_msg*)(msg->data);
  if (!VAR(pending)) {
    VAR(pending) = 1;
    VAR(sentThroughBuffer) = msg;
    VAR(sentMsg) = NULL;
    q->pkt_seq = VAR(max_seq); VAR(max_seq)++;
    InCache(q->pkt_seq,q->pkt_src,q->prev_hop); //save outgoing pkt in cache
    TOS_CALL_COMMAND(DIFFUSION_SUB_SEND_MSG)(TOS_BCAST_ADDR, pkt_type, msg);
    return 1;
  }
  return 0;
}

/*
  If Agent wants the packet to be processed by Diffusion, it returns NULL

  Similarly, if there is no local agent that wishes to process the packet,
  it is handled by Diffusion

  If Agent sends out a new packet and does not want it to be processed by Diffusion
  it returns new buffer.
*/

TOS_MsgPtr Demux(TOS_MsgPtr msg, unsigned char data_type, unsigned char pkt_type) {
  unsigned char x=SENTINEL;

  CLR_RED_LED_PIN();

  if ((x = VAR(port_reg[data_type])) != SENTINEL) {
    if (x == 0) return TOS_SIGNAL_EVENT(DIFFUSION_AGENTS_PORT_0)(msg, pkt_type);
    if (x == 1) return TOS_SIGNAL_EVENT(DIFFUSION_AGENTS_PORT_1)(msg, pkt_type);
    if (x == 2) return TOS_SIGNAL_EVENT(DIFFUSION_AGENTS_PORT_2)(msg, pkt_type);
    if (x == 3) return TOS_SIGNAL_EVENT(DIFFUSION_AGENTS_PORT_3)(msg, pkt_type);
    if (x == 4) return TOS_SIGNAL_EVENT(DIFFUSION_AGENTS_PORT_4)(msg, pkt_type);
    if (x == 5) return TOS_SIGNAL_EVENT(DIFFUSION_AGENTS_PORT_5)(msg, pkt_type);
    if (x == 6) return TOS_SIGNAL_EVENT(DIFFUSION_AGENTS_PORT_6)(msg, pkt_type);
    if (x == 7) return TOS_SIGNAL_EVENT(DIFFUSION_AGENTS_PORT_7)(msg, pkt_type);
  }
 
  return NULL;
}


TOS_MsgPtr ProcessPacket(TOS_MsgPtr msg, char pkt_type, char local_pkt_src) {
  
  query_msg* q=(query_msg*)(msg->data);
  unsigned char i=0;
  char send_msg_type=0;
  TOS_MsgPtr tmp;

  if (local_pkt_src == REMOTE_DIFFUSION_PKT) {
    // if recvd packet before, then discard
    if (InCache(q->pkt_seq,q->pkt_src,q->prev_hop)) return msg;
    
    // Check if packet is intended for a local agent. If so return. Else process further
    if ((tmp = Demux(msg, q->data_type, pkt_type)) != NULL) return tmp;
  }

  switch (pkt_type) {
  case AM_MSG(DIFFUSION_INTEREST_MSG): 
    printf("RECVD INTEREST PACKET: Data_Sink:%d Pkt_Seq:%d Pkt_Src:%d\n",q->data_sink,q->pkt_seq,q->pkt_src);fflush(stdout);
    
    //    CLR_RED_LED_PIN();CLR_GREEN_LED_PIN();CLR_YELLOW_LED_PIN();
    
    if ((i=AddGradient(q->data_type, q->pkt_seq, q->data_sink, q->prev_hop)) == SENTINEL)
      return msg;
    
    // If in NULL_STATE move to INTEREST_STATE
    if (GetState(i)==NULL_STATE)
      SetState(i,INTEREST_STATE);
    
    // forward interest
    if (!VAR(pending) && !VAR(msg_buffer_pending)) {
      // Change prev hop and broadcast
      ((query_msg*)msg->data)->prev_hop = VAR(my_location);
      send_msg_type = AM_MSG(DIFFUSION_INTEREST_MSG);
    }
    break;
    
  case AM_MSG(DIFFUSION_REINFORCEMENT_MSG): 
    printf("RECVD REINFORCEMENT PACKET: Data_Src:%d Data_Sink:%d ,Pkt_Seq:%d Pkt_Src:%d PrevHop=%d NextHop=%d Reinf_Src=%d Reinf_Seq=%d\n",q->data_src,q->data_sink,q->pkt_seq,q->pkt_src,q->prev_hop,q->next_hop,q->reinf_pkt_src,q->reinf_pkt_seq);fflush(stdout);
    
    // If reinforcement is not intended for me dump
    if (q->next_hop != VAR(my_location)) return msg;
    
    //    CLR_GREEN_LED_PIN();
    
    // Find the gradient_list entry
    if ((i = GetGradientEntry(q->data_type, q->data_sink))==SENTINEL) return msg;
    
    // Find the reinf_nbr entry corresponding to data_src
    if (AddReinfNbr(i, q->reinf_pkt_src, q->prev_hop) == SENTINEL) return msg;
    
    // If I am the destined reinforced nbr
    // If there is already a packet in the Xmit Q, the reinforcement will be dropped.
    if (!VAR(pending) && !VAR(msg_buffer_pending)) {
      q->prev_hop = VAR(my_location);
      // Change next hop, prev hop and send it out to our best data source
      
      // Note that LOW DATA RATE was drawn from data_src, therefore reverse parameters
      q->next_hop = GetCacheEntry(q->reinf_pkt_seq, q->reinf_pkt_src);

      //Reinforced packet is generated 
      q->pkt_seq = VAR(max_seq); VAR(max_seq)++; //Increment Seq       

      send_msg_type = AM_MSG(DIFFUSION_REINFORCEMENT_MSG);
    }
    break;
  case AM_MSG(DIFFUSION_LOW_DATA_RATE_MSG): 
    //    CLR_YELLOW_LED_PIN();
    
    printf("RECVD LOW DATA RATE PACKET: Data_Src:%d Data_Sink:%d Pkt_Seq:%d Pkt_Src:%d PrevHop=%d\n",q->data_src,q->data_sink,q->pkt_seq,q->pkt_src,q->prev_hop);fflush(stdout);
    
    // If packet is intended for me
    if (q->data_sink == VAR(my_location)) {
      printf("QUERY REPLY: Light State at %d = %d \n",q->data_src, q->result);
      
      if (!VAR(pending) && !VAR(msg_buffer_pending)) {
	// Set packet being reinforced so that nodes can locate it in the data cache
	q->reinf_pkt_seq = q->pkt_seq;
	q->reinf_pkt_src = q->pkt_src;
	// Set new pkt_src
	q->pkt_src = VAR(my_location);
	// Set next hop as the hop from which the pkt was rcvd
	q->next_hop = q->prev_hop;
	q->prev_hop = VAR(my_location);
	q->pkt_seq = VAR(max_seq); VAR(max_seq)++; //Increment Seq 	
	//	CLR_GREEN_LED_PIN();    
	
	send_msg_type = AM_MSG(DIFFUSION_REINFORCEMENT_MSG);
      }
    } else {
      
      // Modify prev hop info
      q->prev_hop = VAR(my_location);
      
      printf("RECVD New Reading = %d\n",q->result);
      
      // Packet is not intended for me. Rebroadcast.
	if (!VAR(pending) && !VAR(msg_buffer_pending))        
	  send_msg_type = AM_MSG(DIFFUSION_LOW_DATA_RATE_MSG);
    }
    break;
  case AM_MSG(DIFFUSION_HIGH_DATA_RATE_MSG): 
    printf("RECVD HIGH DATA RATE PACKET: Data_Src:%d Data_Sink:%d Pkt_Seq:%d Pkt_Src:%d PrevHop:%d NextHop:%d\n",q->data_src,q->data_sink,q->pkt_seq,q->pkt_src,q->prev_hop,q->next_hop);fflush(stdout);
    
    // If high data rate packet is not intended for me dump
    if (q->next_hop != VAR(my_location)) return msg;
    
    //    CLR_RED_LED_PIN();
    
    printf("RECVD New Reading = %d\n",q->result);
    
    // Packet is intended for me. Send it to node which reinforced the data.
    if (!VAR(pending) && !VAR(msg_buffer_pending)) {
      q->prev_hop = VAR(my_location);
      q->next_hop = GetReinfNbr(GetGradientEntry(q->data_type, q->data_sink), q->data_src);
      q->pkt_seq = VAR(max_seq); VAR(max_seq)++; //Increment Seq      

      send_msg_type = AM_MSG(DIFFUSION_HIGH_DATA_RATE_MSG);
    }
    break;
    }

  if (!VAR(pending) && !VAR(msg_buffer_pending)) {
    switch (local_pkt_src) {
    case REMOTE_DIFFUSION_PKT:
      // Send out msg; Return my buffer.
      VAR(sentMsg) = msg;
      InCache(q->pkt_seq,q->pkt_src,q->prev_hop); //save outgoing pkt in cache
      TOS_CALL_COMMAND(DIFFUSION_SUB_SEND_MSG)(TOS_BCAST_ADDR, send_msg_type, msg);
      VAR(pending) = 1;
      VAR(msg_buffer_pending) = 1;
      tmp = VAR(msg);
      VAR(msg) = msg;
      return tmp;

    case LOCAL_AGENT_PKT:
      // Forward Agent msg.
      q->pkt_seq = VAR(max_seq); VAR(max_seq)++; //Increment Seq
      InCache(q->pkt_seq,q->pkt_src,q->prev_hop); //save outgoing pkt in cache
      TOS_CALL_COMMAND(DIFFUSION_SUB_SEND_MSG)(TOS_BCAST_ADDR, send_msg_type, msg);
      VAR(pending) = 1;
      VAR(msg_buffer_pending) = 1;
      return msg;
    }
  }
  return msg;
}


/* When interest packet arrives
     If packet in cache, discard.
     If the intended recipient is myself, 
        Send low data rate readings;
     If not,
        Forward Packet
*/
TOS_MsgPtr TOS_MSG_EVENT(DIFFUSION_INTEREST_MSG)(TOS_MsgPtr msg) {
  if (VAR(led_state)%2) CLR_GREEN_LED_PIN();
  else SET_GREEN_LED_PIN(); VAR(led_state)++;
  return ProcessPacket(msg, INTEREST_TYPE, REMOTE_DIFFUSION_PKT);
}


/* When a reinforcement is recvd
     For reinforcement, no need to check cache to see if packet rcvd before
     Reject if reinforcement is not intended for me.
     Forward it to nbr that gave best gradient.
*/
TOS_MsgPtr TOS_MSG_EVENT(DIFFUSION_REINFORCEMENT_MSG)(TOS_MsgPtr msg){
  if (VAR(led_state)%2) CLR_GREEN_LED_PIN();
  else SET_GREEN_LED_PIN(); VAR(led_state)++;
  return ProcessPacket(msg, REINFORCEMENT_TYPE, REMOTE_DIFFUSION_PKT);
}


/* When a low data rate packet is recvd
     Check if in cache, if so reject.
     Else Forward
*/

TOS_MsgPtr TOS_MSG_EVENT(DIFFUSION_LOW_DATA_RATE_MSG)(TOS_MsgPtr msg){
  if (VAR(led_state)%2) CLR_GREEN_LED_PIN();
  else SET_GREEN_LED_PIN(); VAR(led_state)++;
  return ProcessPacket(msg, LOW_DATA_RATE_TYPE, REMOTE_DIFFUSION_PKT);
}


/* When a high data rate packet is recvd
     Check if in cache, if so reject.
     If I am next_hop on packet, forward.
     Else drop
*/

TOS_MsgPtr TOS_MSG_EVENT(DIFFUSION_HIGH_DATA_RATE_MSG)(TOS_MsgPtr msg){
  if (VAR(led_state)%2) CLR_GREEN_LED_PIN();
  else SET_GREEN_LED_PIN(); VAR(led_state)++;
  return ProcessPacket(msg, HIGH_DATA_RATE_TYPE, REMOTE_DIFFUSION_PKT);
}
