/*
 *
 * Authors:   Deepak Ganesan
 * History:   created 10/2/2001
 *
 *
 */

/***********

Green LED:

 Set at init
 Blinks for low data rate, interest packet sent

Red LED:
 
 Toggles for Low Data Rate and High Data Rate packets

Yellow LED:

 Blinks when a Reinforcement is sent ???

****************/

#include "tos.h"
#include "SINK.h"

#define NULL_STATE 0
#define PENDING_STATE 1

#define INTEREST_TYPE 102
#define REINFORCEMENT_TYPE 103
#define LOW_DATA_RATE_TYPE 104
#define HIGH_DATA_RATE_TYPE 105

#define SENTINEL 0xff

#define fflush(stdout);

extern const char TOS_LOCAL_ADDRESS;

typedef struct{
  unsigned char data_type;
  unsigned char pkt_seq;
  unsigned char pkt_src;
  unsigned char data_src;
  unsigned char data_sink;
  unsigned char prev_hop;
  unsigned char next_hop;
  unsigned char reinf_pkt_src;
  unsigned char reinf_pkt_seq;
  unsigned char result;
} query_msg;


#define TOS_FRAME_TYPE SINK_frame
TOS_FRAME_BEGIN(SINK_frame) {
  char pending;
  TOS_MsgPtr msg;
  unsigned char max_seq;
  char state;
  char my_location;
  char led_state;
}
TOS_FRAME_END(SINK_frame);


char TOS_COMMAND(SINK_INIT)(){
  int i;

  VAR(pending) = 0;
  VAR(state) = PENDING_STATE;

  VAR(led_state)=0;

  // Associate port 0, Data Type 0 with sink
  TOS_COMMAND(SINK_SUB_DIFFUSION_REGISTER)(0,0);
  
  TOS_COMMAND(SINK_SUB_TIMERS_INIT)();
  printf("SINK initialized\n"); fflush(stdout);
  return 1;
}

char TOS_COMMAND(SINK_START)(){
  return TOS_COMMAND(SINK_SUB_TIMERS_REGISTER)(0,15);
}

void TOS_EVENT(SINK_TIMERS_EVENT)(int port){
  char c;
  printf("Sink Timers Event\n"); fflush(stdout);

  SET_YELLOW_LED_PIN();SET_GREEN_LED_PIN();SET_RED_LED_PIN();
  
  // If interest has timed out, reset state (restart process)
  if (VAR(state) == NULL_STATE) {
    VAR(state) = PENDING_STATE;
    return;
  }

  if (VAR(state) == PENDING_STATE) {
    SINK_OUTPUT(VAR(max_seq));
    VAR(max_seq)++;
  }
}

char SINK_OUTPUT(unsigned char val){
	query_msg* message;
	
	// Get Free Buffer from Diffusion
	if ((VAR(msg) = TOS_COMMAND(SINK_SUB_GET_MSG_BUFFER)()) == NULL) return 0;

	message = (query_msg*)VAR(msg)->data;
	if (!VAR(pending)) {
	  message->pkt_seq = VAR(max_seq);
	  message->pkt_src = VAR(my_location);
	  message->data_sink = VAR(my_location);
	  message->data_type = 0;
	  message->prev_hop = VAR(my_location);
	  message->result = 99;
	  printf("Sending out Seq = %d\n", VAR(max_seq)); fflush(stdout);
	  if (TOS_COMMAND(SINK_SUB_DIFFUSION_SEND_MSG)(VAR(msg), INTEREST_TYPE)) {
	    VAR(pending) = 1;
	    CLR_GREEN_LED_PIN();
	    return 1;
	  }
	}
	return 0;
}


/* Send Completion Event
   Determine if this event was ours.
   If so, process it by releasing buffer to Diffusion layer.
*/
char TOS_EVENT(SINK_SUB_MSG_SEND_DONE)(TOS_MsgPtr sentBuffer){

  if (VAR(pending) && sentBuffer == VAR(msg)) {
    VAR(pending) = 0;
    SET_YELLOW_LED_PIN();SET_GREEN_LED_PIN();SET_RED_LED_PIN();
    TOS_COMMAND(SINK_SUB_DIFFUSION_RELEASE_MSG_BUFFER)(VAR(msg));
    VAR(msg) = NULL;
    return 1;
  }
  return 0;
}


/* Active Message handler
 */

TOS_MsgPtr TOS_MSG_EVENT(SINK_RECEIVE_MSG)(TOS_MsgPtr msg, char pkt_type){
  query_msg *q;

  q = (query_msg*)(msg->data);

  switch (pkt_type) {
  case LOW_DATA_RATE_TYPE:
    
    CLR_RED_LED_PIN();CLR_GREEN_LED_PIN();
    
    printf("RECVD LOW DATA RATE PACKET: Pkt_Seq:%d Pkt_Src:%d PrevHop=%d\n",q->pkt_seq,q->pkt_src,q->prev_hop); fflush(stdout);
    
    /* Query reply obtained. Stop further queries */
    VAR(state) = NULL_STATE;

    printf("QUERY REPLY: Light State at %d = %d \n",q->data_src, q->result); fflush(stdout);
    
    return NULL; //Request Diffusion to further process msg (generate Reinf)

  case HIGH_DATA_RATE_TYPE:

    CLR_RED_LED_PIN();

    /* Query reply obtained. Stop further queries */
    VAR(state) = NULL_STATE;

    printf("RECVD HIGH DATA RATE PACKET: Pkt_Seq:%d Pkt_Src:%d PrevHop=%d\n",q->pkt_seq,q->pkt_src,q->prev_hop); fflush(stdout);

    return msg; //Request Diffusion to dump message without processing
  }
  
  return NULL; //Request Diffusion to process msg
}




