/*
 *
 * Authors:   Deepak Ganesan
 * History:   created 10/2/2001
 *
 *
 */

/***********

When Data is sent: All Leds Blink

When Interest is Received: RED led s/w on

When Reinforcement is Received: RED/Green Leds on
****************/

#include "tos.h"
#include "SRC.h"

#define NULL_STATE 0
#define PENDING_STATE 1

#define INTEREST_TYPE 102
#define REINFORCEMENT_TYPE 103
#define LOW_DATA_RATE_TYPE 104
#define HIGH_DATA_RATE_TYPE 105

#define SENTINEL 0xff

#define fflush(stdout);

extern const char TOS_LOCAL_ADDRESS;

typedef struct{
  unsigned char data_type;
  unsigned char pkt_seq;
  unsigned char pkt_src;
  unsigned char data_src;
  unsigned char data_sink;
  unsigned char prev_hop;
  unsigned char next_hop;
  unsigned char reinf_pkt_src;
  unsigned char reinf_pkt_seq;
  unsigned char result;
} query_msg;


#define TOS_FRAME_TYPE SRC_frame
TOS_FRAME_BEGIN(SRC_frame) {
  char pending;
  TOS_MsgPtr msg;
  unsigned char my_seq;
  unsigned char my_data_type;
  char pkt_type;
  char next_hop;
  char state;
  char my_location;
  char photo_data;              // photo data
  char photo_pending;           // state of photo sensor
}
TOS_FRAME_END(SRC_frame);


char TOS_COMMAND(SRC_INIT)(){
  int i;

  VAR(pending) = 0;
  VAR(photo_pending) = 0;
  printf("addr = %d\n",TOS_LOCAL_ADDRESS);
  VAR(my_location) = TOS_LOCAL_ADDRESS;

  VAR(my_seq) = 0;

  VAR(my_data_type) = 0;

  CLR_GREEN_LED_PIN();
  // Associate port 0, Data Type 0 with sink
  TOS_COMMAND(SRC_SUB_TIMERS_INIT)();
  printf("SRC initialized\n");
  return 1;
}

char TOS_COMMAND(SRC_START)(){
  TOS_COMMAND(SRC_SUB_DIFFUSION_REGISTER)(0,0);
  return TOS_COMMAND(SRC_SUB_TIMERS_REGISTER)(0,8); // PHOTO timer: Port 1
}

void TOS_EVENT(SRC_TIMERS_PHOTO_EVENT)(int port){
  printf("Photo Timer Event\n");fflush(stdout);

  // get photo reading if nothing is pending
  if (!VAR(photo_pending)) {
    TOS_CALL_COMMAND(SRC_GET_PHOTO_DATA)();
    VAR(photo_pending) = 1;
  }
}

/* DATA Timer. Port 2*/
void TOS_EVENT(SRC_TIMERS_DATA_EVENT)(int port){
  query_msg *q;

  printf("Data Timer Event: Port 1\n");fflush(stdout);
  
  if ((VAR(msg) = TOS_COMMAND(SRC_SUB_DIFFUSION_GET_MSG_BUFFER)()) == NULL) return;

  q = (query_msg*)(VAR(msg)->data);
  
  q->pkt_seq = VAR(my_seq); VAR(my_seq)++;
  q->pkt_src = VAR(my_location);
  q->data_type = VAR(my_data_type);
  q->data_src = VAR(my_location);
  q->prev_hop = VAR(my_location);
  q->next_hop = VAR(next_hop);
  printf("SRC: pkt_src=%d\n",q->pkt_src); fflush(stdout);
  VAR(pending) = 1;
  CLR_YELLOW_LED_PIN();CLR_GREEN_LED_PIN();CLR_RED_LED_PIN();
  TOS_COMMAND(SRC_SUB_DIFFUSION_SEND_MSG)(VAR(msg), VAR(pkt_type));
}

/* Photo sensor returns data
   Reset photo_pending
 */
char TOS_EVENT(SRC_GET_PHOTO_DATA_DONE)(int photo_data) {
  VAR(photo_data) = (photo_data >> 7) & 0x7;
  VAR(photo_pending) = 0;
  printf("Photo Reading = %d\n",VAR(photo_data));fflush(stdout);
  return 1;
}

/* Send Completion Event
   Determine if this event was ours.
   If so, process it by releasing buffer to Diffusion layer.
*/
char TOS_EVENT(SRC_SUB_MSG_SEND_DONE)(TOS_MsgPtr sentBuffer){
  printf("In SRC_SUB_MSG_SEND_DONE\n");fflush(stdout);

  if (VAR(pending) && sentBuffer == VAR(msg)) {
    VAR(pending) = 0;
    SET_YELLOW_LED_PIN();SET_GREEN_LED_PIN();SET_RED_LED_PIN();
    TOS_COMMAND(SRC_SUB_DIFFUSION_RELEASE_MSG_BUFFER)(VAR(msg));
    VAR(msg) = NULL;
    return 1;
  }
  return 0;
}


/* Active Message handler
 */

TOS_MsgPtr TOS_MSG_EVENT(SRC_RECEIVE_MSG)(TOS_MsgPtr msg, char pkt_type){
  query_msg *q;

  printf("In SRC_RECEIVE_MSG\n"); fflush(stdout);

  CLR_YELLOW_LED_PIN();

  q = (query_msg*)(msg->data);

  switch (pkt_type) {
  case INTEREST_TYPE:
    CLR_RED_LED_PIN();
    TOS_COMMAND(SRC_SUB_TIMERS_REGISTER)(1,3);
    VAR(pkt_type) = LOW_DATA_RATE_TYPE;
    break;
  case REINFORCEMENT_TYPE:
    if (q->next_hop==VAR(my_location) && q->reinf_pkt_src==VAR(my_location)) {
      CLR_RED_LED_PIN();     CLR_GREEN_LED_PIN();
      TOS_COMMAND(SRC_SUB_TIMERS_REGISTER)(1,1);
      VAR(pkt_type) = HIGH_DATA_RATE_TYPE;
      VAR(next_hop) = q->prev_hop;
      return msg; // Drop msg; No further forwarding of reinforcements
    }
    break;
  }

  return NULL;
}




